"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Base interface for both FollowOne and FollowAll
 */
class Follower {
    preFetch() {
        this.prefetchEnabled = true;
        return this;
    }
}
/**
 * The Follower class is what's being returned from follow() functions.
 *
 * It's 'PromiseLike', which means you can treat it like a Promise, and it
 * can be awaited. When used as a Promise, it resolves to the Resource object
 * that was followed.
 *
 * In addition to being a Promise<Resource> stand-in, it also exposes other
 * functions, namely:
 *
 * * `follow()` to allow a user to chain several follow() functions to do
 *   several 'hops' all at once.
 * * `followAll()`, allowing a user to call `followAll()` at the end of a
 *   chain.
 */
class FollowerOne extends Follower {
    constructor(resource, rel, variables) {
        super();
        this.resource = resource;
        this.rel = rel;
        this.variables = variables;
    }
    /**
     * This 'then' function behaves like a Promise then() function.
     *
     * This method signature is pretty crazy, but trust that it's pretty much
     * like any then() method on a promise.
     */
    then(onfulfilled, onrejected) {
        return this.fetchLinkedResource().then(onfulfilled, onrejected);
    }
    /**
     * This 'then' function behaves like a Promise then() function.
     */
    catch(onrejected) {
        return this.fetchLinkedResource().then(undefined, onrejected);
    }
    /**
     * Follow another link immediately after following this link.
     *
     * This allows you to follow several hops of links in one go.
     *
     * For example: resource.follow('foo').follow('bar');
     */
    follow(rel, variables) {
        return new FollowerOne(this.fetchLinkedResource(), rel, variables);
    }
    /**
     * Follows a set of links immediately after following this link.
     *
     * For example: resource.follow('foo').followAll('item');
     */
    async followAll(rel) {
        return new FollowerMany(this.fetchLinkedResource(), rel);
    }
    /**
     * This function does the actual fetching, to obtained the url
     * of the linked resource. It returns the Resource object.
     */
    async fetchLinkedResource() {
        const resource = await this.resource;
        const link = await resource.link(this.rel);
        let href;
        if (link.templated && this.variables) {
            href = link.expand(this.variables);
        }
        else {
            href = link.resolve();
        }
        const newResource = resource.go(href);
        if (link.type) {
            newResource.contentType = link.type;
        }
        if (this.prefetchEnabled) {
            newResource.get().catch(err => {
                // tslint:disable-next-line no-console
                console.warn('Error while prefetching linked resource', err);
            });
        }
        return newResource;
    }
}
exports.FollowerOne = FollowerOne;
/**
 */
class FollowerMany extends Follower {
    constructor(resource, rel) {
        super();
        this.resource = resource;
        this.rel = rel;
    }
    /**
     * This 'then' function behaves like a Promise then() function.
     */
    then(onfulfilled, onrejected) {
        return this.fetchLinkedResources().then(onfulfilled, onrejected);
    }
    /**
     * This 'then' function behaves like a Promise then() function.
     */
    catch(onrejected) {
        return this.fetchLinkedResources().then(undefined, onrejected);
    }
    /**
     * This function does the actual fetching, to obtained the url
     * of the linked resource. It returns the Resource object.
     */
    async fetchLinkedResources() {
        const resource = await this.resource;
        const links = await resource.links(this.rel);
        let href;
        const result = [];
        for (const link of links) {
            href = link.resolve();
            const newResource = resource.go(href);
            if (link.type) {
                newResource.contentType = link.type;
            }
            result.push(newResource);
            if (this.prefetchEnabled) {
                newResource.get().catch(err => {
                    // tslint:disable-next-line no-console
                    console.warn('Error while prefetching linked resource', err);
                });
            }
        }
        return result;
    }
}
exports.FollowerMany = FollowerMany;
//# sourceMappingURL=follower.js.map