"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const hal_1 = require("./representor/hal");
const html_1 = require("./representor/html");
const jsonapi_1 = require("./representor/jsonapi");
const resource_1 = require("./resource");
const fetch_helper_1 = require("./utils/fetch-helper");
require("./utils/fetch-polyfill");
const url_1 = require("./utils/url");
/**
 * The main Ketting client object.
 */
class Ketting {
    constructor(bookMark, options) {
        if (typeof options === 'undefined') {
            options = {};
        }
        this.resourceCache = {};
        this.contentTypes = [
            {
                mime: 'application/hal+json',
                representor: 'hal',
                q: '1.0',
            },
            {
                mime: 'application/vnd.api+json',
                representor: 'jsonapi',
                q: '0.9',
            },
            {
                mime: 'application/json',
                representor: 'hal',
                q: '0.8',
            },
            {
                mime: 'text/html',
                representor: 'html',
                q: '0.7',
            }
        ];
        this.bookMark = bookMark;
        this.fetchHelper = new fetch_helper_1.default(options, this.beforeRequest.bind(this));
    }
    /**
     * This function is a shortcut for getResource().follow(x);
     */
    follow(rel, variables) {
        return this.getResource().follow(rel, variables);
    }
    /**
     * Returns a resource by its uri.
     *
     * This function doesn't do any HTTP requests. The uri is optional. If it's
     * not specified, it will return the bookmark resource.
     *
     * If a relative uri is passed, it will be resolved based on the bookmark
     * uri.
     */
    go(uri) {
        if (typeof uri === 'undefined') {
            uri = '';
        }
        uri = url_1.resolve(this.bookMark, uri);
        if (!this.resourceCache[uri]) {
            this.resourceCache[uri] = new resource_1.default(this, uri);
        }
        return this.resourceCache[uri];
    }
    /**
     * Returns a resource by its uri.
     *
     * This function doesn't do any HTTP requests. The uri is optional. If it's
     * not specified, it will return the bookmark resource.
     */
    getResource(uri) {
        return this.go(uri);
    }
    /**
     * This function does an arbitrary request using the fetch API.
     *
     * @see {@link https://developer.mozilla.org/en-US/docs/Web/API/GlobalFetch}
     */
    fetch(input, init) {
        return this.fetchHelper.fetch(input, init);
    }
    /**
     * This function returns a representor constructor for a mime type.
     *
     * For example, given text/html, this function might return the constructor
     * stored in representor/html.
     */
    getRepresentor(contentType) {
        if (contentType.indexOf(';') !== -1) {
            contentType = contentType.split(';')[0];
        }
        contentType = contentType.trim();
        const result = this.contentTypes.find(item => {
            return item.mime === contentType;
        });
        if (!result) {
            throw new Error('Could not find a representor for contentType: ' + contentType);
        }
        switch (result.representor) {
            case 'html':
                return html_1.default;
            case 'hal':
                return hal_1.default;
            case 'jsonapi':
                return jsonapi_1.default;
            default:
                throw new Error('Unknown representor: ' + result.representor);
        }
    }
    /**
     * Generates an accept header string, based on registered Resource Types.
     */
    getAcceptHeader() {
        return this.contentTypes
            .map(contentType => {
            let item = contentType.mime;
            if (contentType.q) {
                item += ';q=' + contentType.q;
            }
            return item;
        })
            .join(', ');
    }
    beforeRequest(request) {
        const safeMethods = ['GET', 'HEAD', 'OPTIONS', 'PRI', 'PROPFIND', 'REPORT', 'SEARCH', 'TRACE'];
        if (safeMethods.includes(request.method)) {
            return;
        }
        if (request.url in this.resourceCache) {
            // Clear cache
            this.resourceCache[request.url].clearCache();
        }
    }
}
exports.default = Ketting;
//# sourceMappingURL=ketting.js.map