"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const link_1 = require("../link");
// import { resolve } from '../utils/url';
const base_1 = require("./base");
/**
 * This class represents JSON:API responses.
 *
 * The Representor is responsible from extracting any links from the body,
 * so they can be followed.
 */
class JsonApi extends base_1.default {
    constructor(uri, contentType, body) {
        super(uri, contentType, body);
        if (typeof body === 'string') {
            this.body = JSON.parse(body);
        }
        else {
            this.body = body;
        }
        this.links = [
            ...parseJsonApiLinks(uri, this.body),
            ...parseJsonApiCollection(uri, this.body)
        ];
    }
}
exports.default = JsonApi;
/**
 * This function takes a JSON:API object, and extracts the links property.
 */
function parseJsonApiLinks(baseHref, body) {
    const result = [];
    if (body.links === undefined) {
        return result;
    }
    for (const [rel, linkValue] of Object.entries(body.links)) {
        if (Array.isArray(linkValue)) {
            result.push(...linkValue.map(link => parseJsonApiLink(baseHref, rel, link)));
        }
        else {
            result.push(parseJsonApiLink(baseHref, rel, linkValue));
        }
    }
    return result;
}
/**
 * Find collection members in JSON:API objects.
 *
 * A JSON:API top-level object might represent a collection that has 0 or more
 * members.
 *
 * Members of this collection should appear as an 'item' link to the parent.
 */
function parseJsonApiCollection(baseHref, body) {
    if (!Array.isArray(body.data)) {
        // Not a collection
        return [];
    }
    const result = [];
    for (const member of body.data) {
        if ('self' in member.links) {
            const selfLink = parseJsonApiLink(baseHref, 'self', member.links.self);
            result.push(new link_1.default({
                context: baseHref,
                href: selfLink.href,
                rel: 'item'
            }));
        }
    }
    return result;
}
/**
 * This function takes a single link value from a JSON:API link object, and
 * returns a object of type Link
 */
function parseJsonApiLink(baseHref, rel, link) {
    return new link_1.default({
        context: baseHref,
        rel,
        href: typeof link === 'string' ? link : link.href,
    });
}
//# sourceMappingURL=jsonapi.js.map