import Link from '../link';
import { resolve } from '../utils/url';
import Representation from './base';

/**
 * The Representation class is basically a 'body' of a request
 * or response.
 *
 * This class is for HAL JSON responses.
 */
export default class Hal extends Representation {

  body: { [s: string]: any };

  constructor(uri: string, contentType: string, body: any) {

    super(uri, contentType, body);

    if (typeof body === 'string') {
      this.body = JSON.parse(body);
    } else {
      this.body = body;
    }

    if (typeof this.body._links !== 'undefined') {
      parseHalLinks(this);
    }
    if (typeof this.body._embedded !== 'undefined') {
      parseHalEmbedded(this);
    }

    delete this.body._links;
    delete this.body._embedded;

  }

}

/**
 * Parse the Hal _links object and populate the 'links' property.
 */
const parseHalLinks = (representation: Hal): void => {

  for (const relType of Object.keys((<any> representation.body)._links)) {

    let links = (<any> representation.body)._links[relType];
    if (!Array.isArray(links)) {
      links = [links];
    }
    parseHalLink(representation, relType, links);

  }

};

type HalLink = {
  href: string,
  name?: string,
  title: string,
  templated?: boolean,
  type?: string
};

/**
 * Parses a single HAL link from a _links object, or a list of links.
 */
const parseHalLink = (representation: Hal, rel: string, links: HalLink[]): void => {

  for (const link of links) {
    representation.links.push(
      new Link({
        rel: rel,
        context: representation.uri,
        href: link.href,
        title: link.title,
        type: link.type,
        templated: link.templated,
        name: link.name
      })
    );
  }

};

/**
 * Parse the HAL _embedded object. Right now we're just grabbing the
 * information from _embedded and turn it into links.
 */
const parseHalEmbedded = (representation: Hal): void => {

  for (const relType of Object.keys((<any> representation).body._embedded)) {

    let embedded = (<any> representation).body._embedded[relType];
    if (!Array.isArray(embedded)) {
      embedded = [embedded];
    }
    for (const embeddedItem of embedded) {

      const uri = resolve(
        representation.uri,
        embeddedItem._links.self.href
      );

      // Only adding a link to the representation if it didn't already exist.
      if (!representation.links.find( item => {
        return item.rel === relType && embeddedItem._links.self.href === item.href;
      })) {
        representation.links.push(
          new Link({
            rel: relType,
            context: representation.uri,
            href: embeddedItem._links.self.href
          })
        );
      }

      representation.embedded[uri] = embeddedItem;

    }
  }
};
